@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * Copyright (C) 2003-2007 Imendio AB
- * Copyright (C) 2007-2008 Collabora Ltd.
+ * Copyright (C) 2007-2010 Collabora Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -51,7 +51,7 @@
 #include <libempathy-gtk/empathy-log-window.h>
 #include <libempathy-gtk/empathy-geometry.h>
 #include <libempathy-gtk/empathy-smiley-manager.h>
-#include <libempathy-gtk/empathy-sound.h>
+#include <libempathy-gtk/empathy-sound-manager.h>
 #include <libempathy-gtk/empathy-ui-utils.h>
 #include <libempathy-gtk/empathy-notify-manager.h>
 
@@ -111,13 +111,19 @@ typedef struct {
 
 	/* Last user action time we acted upon to show a tab */
 	guint32    x_user_action_time;
+
+	GSettings *gsettings_chat;
+	GSettings *gsettings_notif;
+	GSettings *gsettings_ui;
+
+	EmpathySoundManager *sound_mgr;
 } EmpathyChatWindowPriv;
 
 static GList *chat_windows = NULL;
 
 static const guint tab_accel_keys[] = {
-	GDK_1, GDK_2, GDK_3, GDK_4, GDK_5,
-	GDK_6, GDK_7, GDK_8, GDK_9, GDK_0
+	GDK_KEY_1, GDK_KEY_2, GDK_KEY_3, GDK_KEY_4, GDK_KEY_5,
+	GDK_KEY_6, GDK_KEY_7, GDK_KEY_8, GDK_KEY_9, GDK_KEY_0
 };
 
 typedef enum {
@@ -146,6 +152,20 @@ static const GtkTargetEntry drag_types_dest_file[] = {
 
 static void chat_window_update (EmpathyChatWindow *window);
 
+static void empathy_chat_window_add_chat (EmpathyChatWindow *window,
+			      EmpathyChat	*chat);
+
+static void empathy_chat_window_remove_chat (EmpathyChatWindow *window,
+				 EmpathyChat	   *chat);
+
+static void empathy_chat_window_move_chat (EmpathyChatWindow *old_window,
+			       EmpathyChatWindow *new_window,
+			       EmpathyChat       *chat);
+
+static void empathy_chat_window_get_nb_chats (EmpathyChatWindow *self,
+			       guint *nb_rooms,
+			       guint *nb_private);
+
 G_DEFINE_TYPE (EmpathyChatWindow, empathy_chat_window, G_TYPE_OBJECT);
 
 static void
@@ -403,10 +423,6 @@ chat_window_contact_menu_update (EmpathyChatWindowPriv *priv,
 
 	if (orig_submenu == NULL || !gtk_widget_get_visible (orig_submenu)) {
 		submenu = empathy_chat_get_contact_menu (priv->current_chat);
-
-		/* gtk_menu_attach_to_widget() doesn't behave nicely here */
-		g_object_set_data (G_OBJECT (submenu), "window", priv->dialog);
-
 		gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu), submenu);
 		gtk_widget_show (menu);
 	} else {
@@ -523,9 +539,7 @@ chat_window_icon_update (EmpathyChatWindowPriv *priv)
 		gtk_window_set_icon_name (GTK_WINDOW (priv->dialog),
 					  EMPATHY_IMAGE_MESSAGE);
 	} else {
-		GSettings *gsettings = g_settings_new (EMPATHY_PREFS_CHAT_SCHEMA);
-
-		avatar_in_icon = g_settings_get_boolean (gsettings,
+		avatar_in_icon = g_settings_get_boolean (priv->gsettings_chat,
 				EMPATHY_PREFS_CHAT_AVATAR_IN_ICON);
 
 		if (n_chats == 1 && avatar_in_icon) {
@@ -539,8 +553,6 @@ chat_window_icon_update (EmpathyChatWindowPriv *priv)
 		} else {
 			gtk_window_set_icon_name (GTK_WINDOW (priv->dialog), NULL);
 		}
-
-		g_object_unref (gsettings);
 	}
 }
 
@@ -1158,6 +1170,12 @@ chat_window_tabs_right_activate_cb (GtkAction         *action,
 	chat_window_menu_context_update (priv, num_pages);
 }
 
+static EmpathyChatWindow *
+empathy_chat_window_new (void)
+{
+	return EMPATHY_CHAT_WINDOW (g_object_new (EMPATHY_TYPE_CHAT_WINDOW, NULL));
+}
+
 static void
 chat_window_detach_activate_cb (GtkAction         *action,
 				EmpathyChatWindow *window)
@@ -1271,14 +1289,9 @@ chat_window_show_or_update_notification (EmpathyChatWindow *window,
 	if (!empathy_notify_manager_notification_is_enabled (priv->notify_mgr)) {
 		return;
 	} else {
-		GSettings *gsettings = g_settings_new (
-				EMPATHY_PREFS_NOTIFICATIONS_SCHEMA);
-
-		res = g_settings_get_boolean (gsettings,
+		res = g_settings_get_boolean (priv->gsettings_notif,
 				EMPATHY_PREFS_NOTIFICATIONS_FOCUS);
 
-		g_object_unref (gsettings);
-
 		if (!res) {
 			return;
 		}
@@ -1352,6 +1365,21 @@ chat_window_set_highlight_room_tab_label (EmpathyChat *chat)
 	g_free (markup);
 }
 
+static gboolean
+empathy_chat_window_has_focus (EmpathyChatWindow *window)
+{
+	EmpathyChatWindowPriv *priv;
+	gboolean              has_focus;
+
+	g_return_val_if_fail (EMPATHY_IS_CHAT_WINDOW (window), FALSE);
+
+	priv = GET_PRIV (window);
+
+	g_object_get (priv->dialog, "has-toplevel-focus", &has_focus, NULL);
+
+	return has_focus;
+}
+
 static void
 chat_window_new_message_cb (EmpathyChat       *chat,
 			    EmpathyMessage    *message,
@@ -1377,7 +1405,7 @@ chat_window_new_message_cb (EmpathyChat       *chat,
 	sender = empathy_message_get_sender (message);
 
 	if (empathy_contact_is_user (sender)) {
-		empathy_sound_play (GTK_WIDGET (priv->dialog),
+		empathy_sound_manager_play (priv->sound_mgr, GTK_WIDGET (priv->dialog),
 				    EMPATHY_SOUND_MESSAGE_OUTGOING);
 	}
 
@@ -1428,7 +1456,7 @@ chat_window_new_message_cb (EmpathyChat       *chat,
 			chat_window_set_highlight_room_tab_label (chat);
 		}
 
-		empathy_sound_play (GTK_WIDGET (priv->dialog),
+		empathy_sound_manager_play (priv->sound_mgr, GTK_WIDGET (priv->dialog),
 		    EMPATHY_SOUND_MESSAGE_INCOMING);
 
 		/* Pending messages have already been displayed in the approver, so we don't
@@ -1443,7 +1471,7 @@ chat_window_new_message_cb (EmpathyChat       *chat,
 }
 
 static GtkNotebook *
-chat_window_detach_hook (GtkNotebook *source,
+notebook_create_window_cb (GtkNotebook *source,
 			 GtkWidget   *page,
 			 gint         x,
 			 gint         y,
@@ -1849,6 +1877,10 @@ chat_window_finalize (GObject *object)
 	g_object_unref (priv->ui_manager);
 	g_object_unref (priv->chatroom_manager);
 	g_object_unref (priv->notify_mgr);
+	g_object_unref (priv->gsettings_chat);
+	g_object_unref (priv->gsettings_notif);
+	g_object_unref (priv->gsettings_ui);
+	g_object_unref (priv->sound_mgr);
 
 	if (priv->notification != NULL) {
 		notify_notification_close (priv->notification, NULL);
@@ -1893,8 +1925,6 @@ empathy_chat_window_class_init (EmpathyChatWindowClass *klass)
 		"  ythickness = 0\n"
 		"}\n"
 		"widget \"*.empathy-close-button\" style \"empathy-close-button-style\"");
-
-	gtk_notebook_set_window_creation_hook (chat_window_detach_hook, NULL, NULL);
 }
 
 static void
@@ -1961,10 +1991,20 @@ empathy_chat_window_init (EmpathyChatWindow *window)
 	g_object_ref (priv->ui_manager);
 	g_object_unref (gui);
 
+	priv->gsettings_chat = g_settings_new (EMPATHY_PREFS_CHAT_SCHEMA);
+	priv->gsettings_notif = g_settings_new (EMPATHY_PREFS_NOTIFICATIONS_SCHEMA);
+	priv->gsettings_ui = g_settings_new (EMPATHY_PREFS_UI_SCHEMA);
 	priv->chatroom_manager = empathy_chatroom_manager_dup_singleton (NULL);
 
+	priv->sound_mgr = empathy_sound_manager_dup_singleton ();
+
 	priv->notebook = gtk_notebook_new ();
-	gtk_notebook_set_group (GTK_NOTEBOOK (priv->notebook), "EmpathyChatWindow");
+
+	g_signal_connect (priv->notebook, "create-window",
+		G_CALLBACK (notebook_create_window_cb), window);
+
+	gtk_notebook_set_group_name (GTK_NOTEBOOK (priv->notebook),
+		"EmpathyChatWindow");
 	gtk_notebook_set_scrollable (GTK_NOTEBOOK (priv->notebook), TRUE);
 	gtk_notebook_popup_enable (GTK_NOTEBOOK (priv->notebook));
 	gtk_box_pack_start (GTK_BOX (chat_vbox), priv->notebook, TRUE, TRUE, 0);
@@ -2062,26 +2102,31 @@ empathy_chat_window_init (EmpathyChatWindow *window)
 
 	priv->chat_manager = empathy_chat_manager_dup_singleton ();
 	priv->chat_manager_chats_changed_id =
-		g_signal_connect (priv->chat_manager, "chats-changed",
+		g_signal_connect (priv->chat_manager, "closed-chats-changed",
 				  G_CALLBACK (chat_window_chat_manager_chats_changed_cb),
 				  window);
 
 	chat_window_chat_manager_chats_changed_cb (priv->chat_manager,
-						   empathy_chat_manager_get_num_chats (priv->chat_manager),
+						   empathy_chat_manager_get_num_closed_chats (priv->chat_manager),
 						   window);
 }
 
-EmpathyChatWindow *
-empathy_chat_window_new (void)
+static GtkWidget *
+empathy_chat_window_get_dialog (EmpathyChatWindow *window)
 {
-	return EMPATHY_CHAT_WINDOW (g_object_new (EMPATHY_TYPE_CHAT_WINDOW, NULL));
+	EmpathyChatWindowPriv *priv;
+
+	g_return_val_if_fail (window != NULL, NULL);
+
+	priv = GET_PRIV (window);
+
+	return priv->dialog;
 }
 
-/* Returns the window to open a new tab in if there is only one window
- * visble, otherwise, returns NULL indicating that a new window should
- * be added.
+/* Returns the window to open a new tab in if there is a suitable window,
+ * otherwise, returns NULL indicating that a new window should be added.
  */
-EmpathyChatWindow *
+static EmpathyChatWindow *
 empathy_chat_window_get_default (gboolean room)
 {
 	GSettings *gsettings = g_settings_new (EMPATHY_PREFS_UI_SCHEMA);
@@ -2102,44 +2147,33 @@ empathy_chat_window_get_default (gboolean room)
 		EmpathyChatWindowPriv *priv;
 		EmpathyChatWindow *chat_window;
 		GtkWidget         *dialog;
+		guint nb_rooms, nb_private;
 
 		chat_window = l->data;
 		priv = GET_PRIV (chat_window);
 
 		dialog = empathy_chat_window_get_dialog (chat_window);
-		if (empathy_window_get_is_visible (GTK_WINDOW (dialog))) {
-			guint nb_rooms, nb_private;
-			empathy_chat_window_get_nb_chats (chat_window, &nb_rooms, &nb_private);
 
-			/* Skip the window if there aren't any rooms in it */
-			if (room && nb_rooms == 0)
-				continue;
+		empathy_chat_window_get_nb_chats (chat_window, &nb_rooms, &nb_private);
 
-			/* Skip the window if there aren't any 1-1 chats in it */
-			if (!room && nb_private == 0)
-				continue;
+		/* Skip the window if there aren't any rooms in it */
+		if (room && nb_rooms == 0)
+			continue;
 
-			/* Found a visible window on this desktop */
-			return chat_window;
-		}
+		/* Skip the window if there aren't any 1-1 chats in it */
+		if (!room && nb_private == 0)
+			continue;
+
+		/* Found a window on this desktop, make it visible if necessary */
+		if (!empathy_window_get_is_visible (GTK_WINDOW (dialog)))
+			empathy_window_present (GTK_WINDOW (dialog));
+		return chat_window;
 	}
 
 	return NULL;
 }
 
-GtkWidget *
-empathy_chat_window_get_dialog (EmpathyChatWindow *window)
-{
-	EmpathyChatWindowPriv *priv;
-
-	g_return_val_if_fail (window != NULL, NULL);
-
-	priv = GET_PRIV (window);
-
-	return priv->dialog;
-}
-
-void
+static void
 empathy_chat_window_add_chat (EmpathyChatWindow *window,
 			      EmpathyChat	*chat)
 {
@@ -2159,21 +2193,35 @@ empathy_chat_window_add_chat (EmpathyChatWindow *window,
 
 	/* If this window has just been created, position it */
 	if (priv->chats == NULL) {
-		GSettings *gsettings = g_settings_new (EMPATHY_PREFS_UI_SCHEMA);
 		const gchar *name = "chat-window";
 		gboolean     separate_windows;
 
-		separate_windows = g_settings_get_boolean (gsettings,
+		separate_windows = g_settings_get_boolean (priv->gsettings_ui,
 				EMPATHY_PREFS_UI_SEPARATE_CHAT_WINDOWS);
 
-		g_object_unref (gsettings);
+		if (empathy_chat_is_room (chat))
+			name = "room-window";
 
 		if (separate_windows) {
+			gint x, y;
+
+			/* Save current position of the window */
+			gtk_window_get_position (GTK_WINDOW (priv->dialog), &x, &y);
+
+			/* First bind to the 'generic' name. So new window for which we didn't
+			* save a geometry yet will have the geometry of the last saved
+			* window (bgo #601191). */
+			empathy_geometry_bind (GTK_WINDOW (priv->dialog), name);
+
+			/* Restore previous position of the window so the newly created window
+			* won't be in the same position as the latest saved window and so
+			* completely hide it. */
+			gtk_window_move (GTK_WINDOW (priv->dialog), x, y);
+
+			/* Then bind it to the name of the contact/room so we'll save the
+			* geometry specific to this window */
 			name = empathy_chat_get_id (chat);
 		}
-		else if (empathy_chat_is_room (chat)) {
-			name = "room-window";
-		}
 
 		empathy_geometry_bind (GTK_WINDOW (priv->dialog), name);
 	}
@@ -2208,7 +2256,7 @@ empathy_chat_window_add_chat (EmpathyChatWindow *window,
 	DEBUG ("Chat added (%d references)", G_OBJECT (chat)->ref_count);
 }
 
-void
+static void
 empathy_chat_window_remove_chat (EmpathyChatWindow *window,
 				 EmpathyChat	   *chat)
 {
@@ -2246,7 +2294,7 @@ empathy_chat_window_remove_chat (EmpathyChatWindow *window,
 	g_object_unref (chat);
 }
 
-void
+static void
 empathy_chat_window_move_chat (EmpathyChatWindow *old_window,
 			       EmpathyChatWindow *new_window,
 			       EmpathyChat       *chat)
@@ -2275,7 +2323,7 @@ empathy_chat_window_move_chat (EmpathyChatWindow *old_window,
 	g_object_unref (chat);
 }
 
-void
+static void
 empathy_chat_window_switch_to_chat (EmpathyChatWindow *window,
 				    EmpathyChat	      *chat)
 {
@@ -2293,21 +2341,6 @@ empathy_chat_window_switch_to_chat (EmpathyChatWindow *window,
 				       page_num);
 }
 
-gboolean
-empathy_chat_window_has_focus (EmpathyChatWindow *window)
-{
-	EmpathyChatWindowPriv *priv;
-	gboolean              has_focus;
-
-	g_return_val_if_fail (EMPATHY_IS_CHAT_WINDOW (window), FALSE);
-
-	priv = GET_PRIV (window);
-
-	g_object_get (priv->dialog, "has-toplevel-focus", &has_focus, NULL);
-
-	return has_focus;
-}
-
 EmpathyChat *
 empathy_chat_window_find_chat (TpAccount   *account,
 			       const gchar *id)
@@ -2389,7 +2422,7 @@ empathy_chat_window_present_chat (EmpathyChat *chat,
 	gtk_widget_grab_focus (chat->input_text_view);
 }
 
-void
+static void
 empathy_chat_window_get_nb_chats (EmpathyChatWindow *self,
 			       guint *nb_rooms,
 			       guint *nb_private)
